from sklearn.cluster import MeanShift, estimate_bandwidth
import numpy as np
import matplotlib.pyplot as plt
from itertools import cycle
import sys

# upload point cloud data
np.set_printoptions(suppress=True)
filename = "\\point cloud data.txt"
data = np.loadtxt(filename, delimiter=' ')[:, 0:3]
## select data that higher than 2 meter
data = data[data[:,2]>2]
print("data\n",data)
X = data[:, 0:2]
print("X\n",X)

# Mean shift
## The following bandwidth can be automatically detected using
bandwidth = 2.5
ms = MeanShift(bandwidth=bandwidth, bin_seeding=True, max_iter=300, cluster_all=True)
ms.fit(X)

labels = ms.labels_
print("labels\n",labels)
cluster_centers = ms.cluster_centers_
print("cluster_centers\n",cluster_centers)
labels_unique = np.unique(labels)
print("labels_unique\n",labels_unique)
n_clusters_ = len(set(labels)) - (1 if -1 in labels else 0)
print("number of estimated clusters : %d" % n_clusters_)
plt.figure(1)
plt.clf()
colors = cycle("bgrcmy")

# Viusalization
for k, col in zip(range(n_clusters_), colors):
    my_members = labels == k
    cluster_center = cluster_centers[k]
    plt.plot(data[my_members, 0], data[my_members, 1], col + "+")
    plt.plot(
        cluster_center[0],
        cluster_center[1],
        "o",
        markerfacecolor=col,
        markeredgecolor="k",
        markersize=5,
    )
plt.title("Estimated number of clusters: %d" % n_clusters_)
plt.show()

# output the result
x = data[:,0]
y = data[:,1]
z = data[:,2]
newlabels = []
for i in labels:
    newlabels.append(i+1)

datanew = [x, y, z, newlabels]
datanew = np.array(datanew)
datanew = np.transpose(datanew)

## output result
np.set_printoptions(threshold=10000000000000)
outputfile = open("\\point cloud data_MeanShift.txt", "a")
sys.stdout = outputfile
print(datanew)