import numpy as np
from sklearn.cluster import DBSCAN
from sklearn.neighbors import KNeighborsClassifier
import cv2
import time
import sys

def PCA(data, correlation=False, sort=True):
    mean_data = np.mean(data, axis=0)
    normal_data = data - mean_data
    H = np.dot(normal_data.T, normal_data)
    eigenvectors, eigenvalues, _ = np.linalg.svd(H)
    if sort:
        sort = eigenvalues.argsort()[::-1]
        eigenvalues = eigenvalues[sort]
        eigenvectors = eigenvectors[:, sort]
    return eigenvalues, eigenvectors

def compute_polygon_area(points):
    point_num = len(points)
    if(point_num < 3): return 0.0
    s = points[0][1] * (points[point_num-1][0] - points[1][0])
    for i in range(1, point_num):
        s += points[i][1] * (points[i-1][0] - points[(i+1)%point_num][0])
    return abs(s/2.0)

def reclassify(data, eps):
    TreeID = data[:,3]
    TreeID_unique = np.unique(TreeID)

    Tree_main_x = []
    Tree_main_y = []
    Tree_main_z = []
    Tree_main_id = []
    Tree_noise_x = []
    Tree_noise_y = []
    Tree_noise_z = []

    for i in TreeID_unique:
        oneTree = data[data[:,3] == i]
        oneTree = np.array(oneTree)
        oneTree_XY = oneTree[:, 0:2]
        oneTree_XY = np.array(oneTree_XY)

        # Compute DBSCAN
        db = DBSCAN(eps=eps, min_samples=20).fit(oneTree_XY)
        core_samples_mask = np.zeros_like(db.labels_, dtype=bool)
        core_samples_mask[db.core_sample_indices_] = True
        labels = db.labels_
        unique_labels = np.unique(labels)

        # PCA
        # # Caluculate 𝑁𝑖, 𝐴𝑖, 𝑆𝑖
        Tree_shape = []
        length = []
        area = []

        for i in unique_labels:
            if i == -1:
                continue
            eigenvalues, eigenvectors = PCA(oneTree_XY[labels == i])
            value1 = (eigenvalues[0])
            value2 = (eigenvalues[1])

            Tree_shape.append(value1/(value1 + value2))
            length.append(len(labels[labels == i]))

            test_onetreeXY = np.trunc(oneTree_XY[labels == i])
            test_onetreeXY = test_onetreeXY.astype(int)
            hull = cv2.convexHull(test_onetreeXY)
            area_i = compute_polygon_area(hull[:, 0])
            area.append(area_i)

        ## Threshold of three indicators
        length_th = np.sum(length)*0.3
        area_th = np.sum(area)*0.3
        shape_th = []

        for j in range(len(area)):
            area_j = area[j]
            length_j = length[j]
            if length_j < length_th:
                continue
            if area_j < area_th:
                continue
            shape_th.append(Tree_shape[j])

        if shape_th == []:
            continue
        min_shape = min(shape_th)
        index = Tree_shape.index(min_shape)

        if unique_labels[0] == -1:
            num = unique_labels[index+1]
        else:
            num = unique_labels[index]

        ## denfine the core tree and noise points
        oneTree_main = oneTree[labels == num]
        oneTree_noise = oneTree[labels != num]
        for j in range(len(oneTree_main)):
            Tree_main_x.append(oneTree_main[j][0])
            Tree_main_y.append(oneTree_main[j][1])
            Tree_main_z.append(oneTree_main[j][2])
            Tree_main_id.append(oneTree_main[j][3])
        for j in range(len(oneTree_noise)):
            Tree_noise_x.append(oneTree_noise[j][0])
            Tree_noise_y.append(oneTree_noise[j][1])
            Tree_noise_z.append(oneTree_noise[j][2])

    Tree_main = [Tree_main_x,Tree_main_y,Tree_main_z,Tree_main_id]
    Tree_main = np.transpose(Tree_main)
    Tree_main = np.array(Tree_main)
    Tree_noise = [Tree_noise_x,Tree_noise_y,Tree_noise_z]
    Tree_noise = np.transpose(Tree_noise)
    Tree_noise = np.array(Tree_noise)
    print("Tree_main\n",Tree_main)
    print("Tree_noise\n",Tree_noise)

    ## KNN classifier
    TreeID = Tree_main[:,3]
    XYZ = Tree_main[:,0:3]
    knn = KNeighborsClassifier(n_neighbors=3)
    knn.fit(XYZ, TreeID)
    y_test = knn.predict(Tree_noise)

    Tree_noise_result = [Tree_noise[:,0], Tree_noise[:,1], Tree_noise[:,2], y_test]
    Tree_noise_result = np.transpose(Tree_noise_result)
    Tree_noise_result = np.array(Tree_noise_result)

    newdata = np.vstack((Tree_main,Tree_noise_result))
    return newdata

np.set_printoptions(suppress=True)
filename = "\\point cloud data.txt"
data = np.loadtxt(filename, delimiter=' ')[:, 0:4]
data = data[data[:,2]>5]
print("data\n",data)

# parameters of eps
start_time = time.time()
newdata  = reclassify(data, 0.4)
newdata = reclassify(newdata , 0.55)
newdata = reclassify(newdata, 0.7)

# Calculate process time
end_time = time.time()
time_process = end_time - start_time
print("time_process\n",time_process)

# output the result
np.set_printoptions(threshold=10000000000000)
outputfile = open("\\point cloud data_refining.txt", "a")
sys.stdout = outputfile
print(newdata)